"""
Really long regex code presets:<br>
`quotes` match everything inside quotes<br>
`brackets` match everything inside brackets<br>
`outside_strings` match the pattern but not inside strings<br>
`all_except` match everything except a certain pattern<br>
`func_call` match a function call<br>
`rgba_str` match an RGBA color<br>
`hsla_str` match a HSLA color
"""

import re as _re


class Regex:

    @staticmethod
    def quotes() -> str:
        """Matches everything inside quotes. (Strings)\n
        ------------------------------------------------------------------------------------
        Will create two named groups:<br>
        **`quote`** the quote type (single or double)<br>
        **`string`** everything inside the found quote pair\n
        ------------------------------------------------------------------------------------
        **Attention:** Requires non standard library `regex` not standard library `re`!
        """
        return r'(?P<quote>[\'"])(?P<string>(?:\\.|(?!\g<quote>).)*?)\g<quote>'

    @staticmethod
    def brackets(bracket1: str = "(", bracket2: str = ")", is_group: bool = False) -> str:
        """Matches everything inside brackets, including other nested brackets.\n
        ------------------------------------------------------------------------------------
        **Attention:** Requires non standard library `regex` not standard library `re`!
        """
        g, b1, b2 = (
            "" if is_group else "?:",
            _re.escape(bracket1) if len(bracket1) == 1 else bracket1,
            _re.escape(bracket2) if len(bracket2) == 1 else bracket2,
        )
        return rf'{b1}\s*({g}(?:[^{b1}{b2}"\']|"(?:\\.|[^"\\])*"|\'(?:\\.|[^\'\\])*\'|{b1}(?:[^{b1}{b2}"\']|"(?:\\.|[^"\\])*"|\'(?:\\.|[^\'\\])*\'|(?R))*{b2})*)\s*{b2}'

    @staticmethod
    def outside_strings(pattern: str = r".*") -> str:
        """Matches the `pattern` only when it is not found inside a string (`'...'` or `"..."`)."""
        return rf'(?<!["\'])(?:{pattern})(?!["\'])'

    @staticmethod
    def all_except(
        disallowed_pattern: str,
        ignore_pattern: str = "",
        is_group: bool = False,
    ) -> str:
        """Matches everything except `disallowed_pattern`, unless the `disallowed_pattern` is found inside a string (`'...'` or `"..."`).\n
        ----------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        The `ignore_pattern` is just always ignored. For example if `disallowed_pattern` is `>` and `ignore_pattern` is `->`, the `->`-arrows will be allowed, even though they have `>` in them.<br>
        If `is_group` is `True`, you will be able to reference the matched content as a group (e.g. <code>match.group(<i>int</i>)</code> or <code>r'\\<i>int</i>'</code>).
        """
        return rf'({"" if is_group else "?:"}(?:(?!{ignore_pattern}).)*(?:(?!{Regex.outside_strings(disallowed_pattern)}).)*)'

    @staticmethod
    def func_call(func_name: str = None) -> str:
        """Match a function call<br>
        **`1`** function name<br>
        **`2`** the function's arguments\n
        If no `func_name` is given, it will match any function call.\n
        ------------------------------------------------------------------------------------
        **Attention:** Requires non standard library `regex` not standard library `re`!
        """
        return r"(?<=\b)(" + (func_name if func_name else r"[\w_]+") + r")\s*" + Regex.brackets("(", ")", is_group=True)

    @staticmethod
    def rgba_str(fix_sep: str = ",", allow_alpha: bool = True) -> str:
        """Matches an RGBA color inside a string.\n
        --------------------------------------------------------------------------------
        The RGBA color can be in the formats (for `fix_sep = ','`):<br>
        `rgba(r, g, b)`<br>
        `rgba(r, g, b, a)` (if `allow_alpha=True`)<br>
        `(r, g, b)`<br>
        `(r, g, b, a)` (if `allow_alpha=True`)<br>
        `r, g, b`<br>
        `r, g, b, a` (if `allow_alpha=True`)\n
        --------------------------------------------------------------------------------
        ### Valid ranges:<br>
        `r` 0-255 (amount: red)<br>
        `g` 0-255 (amount: green)<br>
        `b` 0-255 (amount: blue)<br>
        `a` 0-1 (float: opacity)\n
        --------------------------------------------------------------------------------
        If the `fix_sep` is set to nothing, any char that is not a letter or number<br>
        can be used to separate the RGBA values, including just a space."""
        if fix_sep in (None, ""):
            fix_sep = r"[^0-9A-Z]"
        else:
            fix_sep = _re.escape(fix_sep)
        rgb_part = rf"""((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))
            (?:\s*{fix_sep}\s*)((?:0*(?:25[0-5]|2[0-4][0-9]|1?[0-9]{{1,2}})))"""
        return (
            rf"""(?ix)
            (?:rgb|rgba)?\s*(?:\(?\s*{rgb_part}
                (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
            \s*\)?)"""
            if allow_alpha
            else rf"(?ix)(?:rgb|rgba)?\s*(?:\(?\s*{rgb_part}\s*\)?)"
        )

    @staticmethod
    def hsla_str(fix_sep: str = ",", allow_alpha: bool = True) -> str:
        """Matches a HSLA color inside a string.\n
        --------------------------------------------------------------------------------
        The HSLA color can be in the formats (for `fix_sep = ','`):<br>
        `hsla(h, s, l)`<br>
        `hsla(h, s, l, a)` (if `allow_alpha=True`)<br>
        `(h, s, l)`<br>
        `(h, s, l, a)` (if `allow_alpha=True`)<br>
        `h, s, l`<br>
        `h, s, l, a` (if `allow_alpha=True`)\n
        --------------------------------------------------------------------------------
        ### Valid ranges:<br>
        `h` 0-360 (degrees: hue)<br>
        `s` 0-100 (percentage: saturation)<br>
        `l` 0-100 (percentage: lightness)<br>
        `a` 0-1 (float: opacity)\n
        --------------------------------------------------------------------------------
        If the `fix_sep` is set to nothing, any char that is not a letter or number<br>
        can be used to separate the HSLA values, including just a space."""
        if fix_sep in (None, ""):
            fix_sep = r"[^0-9A-Z]"
        else:
            fix_sep = _re.escape(fix_sep)
        hsl_part = rf"""((?:0*(?:360|3[0-5][0-9]|[12][0-9][0-9]|[1-9]?[0-9])))
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))
            (?:\s*{fix_sep}\s*)((?:0*(?:100|[1-9][0-9]|[0-9])))"""
        return (
            rf"""(?ix)
            (?:hsl|hsla)?\s*(?:\(?\s*{hsl_part}
                (?:(?:\s*{fix_sep}\s*)((?:0*(?:0?\.[0-9]+|1\.0+|[0-9]+\.[0-9]+|[0-9]+))))?
            \s*\)?)"""
            if allow_alpha
            else rf"(?ix)(?:hsl|hsla)?\s*(?:\(?\s*{hsl_part}\s*\)?)"
        )

    @staticmethod
    def hexa_str(allow_alpha: bool = True) -> str:
        """Matches a HEXA color inside a string.\n
        --------------------------------------------------------------------------
        The HEXA color can be in the formats (prefix `#`, `0x` or no prefix):<br>
        `RGB`<br>
        `RGBA` (if `allow_alpha=True`)<br>
        `RRGGBB`<br>
        `RRGGBBAA` (if `allow_alpha=True`)\n
        --------------------------------------------------------------------------
        ### Valid ranges:<br>
        each channel from 0-9 and A-F (*case insensitive*)"""
        return (
            r"(?i)^(?:#|0x)?[0-9A-F]{8}|[0-9A-F]{6}|[0-9A-F]{4}|[0-9A-F]{3}$"
            if allow_alpha
            else r"(?i)^(?:#|0x)?[0-9A-F]{6}|[0-9A-F]{3}$"
        )



"""
Functions to be able to use special (easy) formatting codes directly inside some message (string).<br>
These codes, when used within following functions, will change the look of log within the console:
- `FormatCodes.print()` (*print a special format-codes containing string*)
- `FormatCodes.input()` (*input with a special format-codes containing prompt*)
- `FormatCodes.to_ansi()` (*transform all special format-codes into ANSI codes in a string*)\n
--------------------------------------------------------------------------------------------------------------------
How to change the text format and color?<br>
**Example string with formatting codes:**<br>
> `[bold]This is bold text, [#F08]which is pink now [black|BG:#FF0088] and now it changed`<br>
> `to black with a pink background. [_]And this is the boring text, where everything is reset.`\n
⇾ **Instead of writing the formats all separate** `[…][…][…]` **you can join them like this** `[…|…|…]`\n
--------------------------------------------------------------------------------------------------------------------
You can also automatically reset a certain format, behind text like shown in the following example:<br>
> `This is normal text [b](which is bold now) but now it was automatically reset to normal.`\n
This will only reset formats, that have a reset listed below. Colors and BG-colors won't be reset.<br>
This is what will happen, if you use it with a color-format:<br>
> `[cyan]This is cyan text [b](which is bold now.) Now it's not bold any more but still cyan.`\n
If you want to ignore the `()` brackets you can put a `\\` or `/` between:<br>
> `[cyan]This is cyan text [b]/(which is bold now.) And now it is still bold and cyan.`\n
⇾ **To see these examples in action, you can put them into the** `FormatCodes.print()` **function.**\n
--------------------------------------------------------------------------------------------------------------------
**All possible formatting codes:**
- HEX colors:  `[#F08]` or `[#FF0088]` (*with or without leading #*)
- RGB colors:  `[rgb(255, 0, 136)]`
- bright colors:  `[bright:#F08]`
- background colors:  `[BG:#F08]`
- standard cmd colors:
    - `[black]`
    - `[red]`
    - `[green]`
    - `[yellow]`
    - `[blue]`
    - `[magenta]`
    - `[cyan]`
    - `[white]`
- bright cmd colors: `[bright:black]` or `[br:black]`, `[bright:red]` or `[br:red]`, ...
- background cmd colors: `[BG:black]`, `[BG:red]`, ...
- bright background cmd colors: `[BG:bright:black]` or `[BG:br:black]`, `[BG:bright:red]` or `[BG:br:red]`, ...<br>
    ⇾ **The order of** `BG:` **and** `bright:` or `br:` **does not matter.**
- text formats:
    - `[bold]` or `[b]`
    - `[dim]`
    - `[italic]` or `[i]`
    - `[underline]` or `[u]`
    - `[inverse]`, `[invert]` or `[in]`
    - `[hidden]`, `[hide]` or `[h]`
    - `[strikethrough]` or `[s]`
    - `[double-underline]` or `[du]`
- specific reset:  `[_bold]` or `[_b]`, `[_dim]`, ... or `[_color]` or `[_c]`, `[_background]` or `[_bg]`
- total reset: `[_]` (only if no `default_color` is set, otherwise see **↓** )
--------------------------------------------------------------------------------------------------------------------
**Special formatting when param `default_color` is set to a color:**
- `[*]` will reset everything, just like `[_]`, but the text-color will remain in `default_color`
- `[*color]` will reset the text-color, just like `[_color]`, but then also make it `default_color`
- `[default]` will just color the text in `default_color`,
- `[BG:default]` will color the background in `default_color`\n
Unlike the standard cmd colors, the default color can be changed by using the following modifiers:
- `[l]` will lighten the `default_color` text by `brightness_steps`%
- `[ll]` will lighten the `default_color` text by `2 × brightness_steps`%
- `[lll]` will lighten the `default_color` text by `3 × brightness_steps`%
- ... etc. Same thing for darkening:
- `[d]` will darken the `default_color` text by `brightness_steps`%
- `[dd]` will darken the `default_color` text by `2 × brightness_steps`%
- `[ddd]` will darken the `default_color` text by `3 × brightness_steps`%
- ... etc.\n
Per default, you can also use `+` and `-` to get lighter and darker `default_color` versions.
"""

from ._consts_ import ANSI
from .xx_string import String
from .xx_regex import Regex
from .xx_color import *

from functools import lru_cache
import ctypes as _ctypes
import regex as _rx
import sys as _sys
import re as _re


COMPILED = {  # PRECOMPILE REGULAR EXPRESSIONS
    "*": _re.compile(r"\[\s*([^]_]*?)\s*\*\s*([^]_]*?)\]"),
    "*color": _re.compile(r"\[\s*([^]_]*?)\s*\*color\s*([^]_]*?)\]"),
    "format": _rx.compile(
        Regex.brackets("[", "]", is_group=True) + r"(?:\s*([/\\]?)\s*" + Regex.brackets("(", ")", is_group=True) + r")?"
    ),
    "bg?_default": _re.compile(r"(?i)((?:BG\s*:)?)\s*default"),
    "bg_default": _re.compile(r"(?i)BG\s*:\s*default"),
    "modifier": _re.compile(
        rf'(?i)((?:BG\s*:)?)\s*({"|".join([f"{_re.escape(m)}+" for m in ANSI.modifier["lighten"] + ANSI.modifier["darken"]])})$'
    ),
    "rgb": _re.compile(r"(?i)^\s*(BG\s*:)?\s*(?:rgb|rgba)?\s*\(?\s*(\d{1,3})\s*,\s*(\d{1,3})\s*,\s*(\d{1,3})\s*\)?\s*$"),
    "hex": _re.compile(r"(?i)^\s*(BG\s*:)?\s*(?:#|0x)?([0-9A-F]{6}|[0-9A-F]{3})\s*$"),
}


class FormatCodes:

    @staticmethod
    def print(
        *values: object,
        default_color: hexa | rgba = None,
        brightness_steps: int = 20,
        sep: str = " ",
        end: str = "\n",
        flush: bool = True,
    ) -> None:
        FormatCodes.__config_console()
        _sys.stdout.write(FormatCodes.to_ansi(sep.join(map(str, values)) + end, default_color, brightness_steps))
        if flush:
            _sys.stdout.flush()

    @staticmethod
    def input(
        prompt: object = "",
        default_color: hexa | rgba = None,
        brightness_steps: int = 20,
    ) -> str:
        FormatCodes.__config_console()
        return input(FormatCodes.to_ansi(prompt, default_color, brightness_steps))

    @staticmethod
    def to_ansi(
        string: str, default_color: hexa | rgba = None, brightness_steps: int = 20, _default_start: bool = True
    ) -> str:
        result = ""
        if Color.is_valid_rgba(default_color, False):
            use_default = True
        elif Color.is_valid_hexa(default_color, False):
            use_default, default_color = True, Color.to_rgba(default_color)
        else:
            use_default = False
        if use_default:
            string = COMPILED["*"].sub(r"[\1_|default\2]", string)  # REPLACE `[…|*|…]` WITH `[…|_|default|…]`
            string = COMPILED["*color"].sub(r"[\1default\2]", string)  # REPLACE `[…|*color|…]` WITH `[…|default|…]`

        def replace_keys(match: _re.Match) -> str:
            format_keys = match.group(1)
            esc = match.group(2)
            auto_reset_txt = match.group(3)
            if not format_keys:
                return match.group(0)
            format_keys = [k.strip() for k in format_keys.split("|") if k.strip()]
            ansi_formats = [FormatCodes.__get_replacement(k, default_color, brightness_steps) for k in format_keys]
            if auto_reset_txt and not esc:
                reset_keys = [
                    (
                        "_color"
                        if k in ANSI.color_map or Color.is_valid_rgba(k) or Color.is_valid_hexa(k)
                        else (
                            "_bg"
                            if ({"bg", "bright", "br"} & set(k.lower().split(":")))
                            and len(k.split(":")) <= 3
                            and any(
                                k[i:] in ANSI.color_map or Color.is_valid_rgba(k[i:]) or Color.is_valid_hexa(k[i:])
                                for i in range(len(k))
                            )
                            else f"_{k}"
                        )
                    )
                    for k in format_keys
                ]
                ansi_resets = [
                    r
                    for k in reset_keys
                    if (r := FormatCodes.__get_replacement(k, default_color, brightness_steps)).startswith(
                        f"{ANSI.char}{ANSI.start}"
                    )
                ]
            else:
                ansi_resets = []
            if not all(f.startswith(f"{ANSI.char}{ANSI.start}") for f in ansi_formats):
                return match.group(0)
            return (
                "".join(ansi_formats)
                + (
                    f"({FormatCodes.to_ansi(auto_reset_txt, default_color, brightness_steps, False)})"
                    if esc and auto_reset_txt
                    else auto_reset_txt if auto_reset_txt else ""
                )
                + ("" if esc else "".join(ansi_resets))
            )

        return (
            (FormatCodes.__get_default_ansi(default_color) if _default_start else "")
            + (result := "\n".join(COMPILED["format"].sub(replace_keys, line) for line in string.split("\n")))
            if use_default
            else result
        )

    @staticmethod
    @lru_cache(maxsize=64)
    def __config_console() -> None:
        _sys.stdout.flush()
        kernel32 = _ctypes.windll.kernel32
        h = kernel32.GetStdHandle(-11)
        mode = _ctypes.c_ulong()
        kernel32.GetConsoleMode(h, _ctypes.byref(mode))
        kernel32.SetConsoleMode(h, mode.value | 0x0004)

    @staticmethod
    def __get_default_ansi(
        default_color: tuple,
        format_key: str = None,
        brightness_steps: int = None,
        _modifiers: tuple[str, str] = (ANSI.modifier["lighten"], ANSI.modifier["darken"]),
    ) -> str | None:
        if not brightness_steps or (format_key and COMPILED["bg?_default"].search(format_key)):
            return (ANSI.seq_bg_color if format_key and COMPILED["bg_default"].search(format_key) else ANSI.seq_color).format(
                *default_color[:3]
            )
        if not (format_key in _modifiers[0] or format_key in _modifiers[1]):
            return None
        match = COMPILED["modifier"].match(format_key)
        if not match:
            return None
        is_bg, modifiers = match.groups()
        adjust = 0
        for mod in _modifiers[0] + _modifiers[1]:
            adjust = String.single_char_repeats(modifiers, mod)
            if adjust and adjust > 0:
                modifiers = mod
                break
        if adjust == 0:
            return None
        elif modifiers in _modifiers[0]:
            new_rgb = Color.adjust_lightness(default_color, (brightness_steps / 100) * adjust)
        elif modifiers in _modifiers[1]:
            new_rgb = Color.adjust_lightness(default_color, -(brightness_steps / 100) * adjust)
        return (ANSI.seq_bg_color if is_bg else ANSI.seq_color).format(*new_rgb[:3])

    @staticmethod
    def __get_replacement(format_key: str, default_color: rgba = None, brightness_steps: int = 20) -> str:
        """Gives you the corresponding ANSI code for the given format key.<br>
        If `default_color` is not `None`, the text color will be `default_color` if all formats<br>
        are reset or you can get lighter or darker version of `default_color` (also as BG)"""
        use_default = default_color and Color.is_valid_rgba(default_color, False)
        _format_key, format_key = format_key, (  # NORMALIZE THE FORMAT KEY (+ SAVE ORIGINAL)
            "bg:" if "bg" in (parts := format_key.replace(" ", "").lower().split(":")) else ""
        ) + ("bright:" if any(x in parts for x in ("bright", "br")) else "") + ":".join(
            p for p in parts if p not in ("bg", "bright", "br")
        )
        if use_default:
            if new_default_color := FormatCodes.__get_default_ansi(default_color, format_key, brightness_steps):
                return new_default_color
        for map_key in ANSI.codes_map:
            if (isinstance(map_key, tuple) and format_key in map_key) or format_key == map_key:
                return ANSI.seq().format(
                    next(
                        (
                            v
                            for k, v in ANSI.codes_map.items()
                            if format_key == k or (isinstance(k, tuple) and format_key in k)
                        ),
                        None,
                    )
                )
        rgb_match = _re.match(COMPILED["rgb"], format_key)
        hex_match = _re.match(COMPILED["hex"], format_key)
        try:
            if rgb_match:
                is_bg = rgb_match.group(1)
                r, g, b = map(int, rgb_match.groups()[1:])
                if Color.is_valid_rgba((r, g, b)):
                    return ANSI.seq_bg_color.format(r, g, b) if is_bg else ANSI.seq_color.format(r, g, b)
            elif hex_match:
                is_bg = hex_match.group(1)
                rgb = Color.to_rgba(hex_match.group(2))
                return (
                    ANSI.seq_bg_color.format(rgb[0], rgb[1], rgb[2])
                    if is_bg
                    else ANSI.seq_color.format(rgb[0], rgb[1], rgb[2])
                )
        except Exception:
            pass
        return _format_key



import subprocess as _subprocess
import platform as _platform
import time as _time
import sys as _sys
import os as _os


class System:

    @staticmethod
    def restart(
        prompt: object = None,
        wait: int = 0,
        continue_program: bool = False,
        force: bool = False,
    ) -> None:
        """Starts a system restart:<br>
        `prompt` is the message to be displayed in the systems restart notification.<br>
        `wait` is the time to wait until restarting in seconds.<br>
        `continue_program` is whether to continue the current Python program after calling this function.<br>
        `force` is whether to force a restart even if other processes are still running.
        """
        system = _platform.system().lower()
        if system == "windows":
            if not force:
                output = _subprocess.check_output("tasklist", shell=True).decode()
                processes = [line.split()[0] for line in output.splitlines()[3:] if line.strip()]
                if len(processes) > 2:  # EXCLUDING THE PYTHON PROCESS AND CMD
                    raise RuntimeError("Processes are still running. Use the parameter `force=True` to restart anyway.")
            if prompt:
                _os.system(f'shutdown /r /t {wait} /c "{prompt}"')
            else:
                _os.system("shutdown /r /t 0")
            if continue_program:
                print(f"Restarting in {wait} seconds...")
                _time.sleep(wait)
        elif system in ("linux", "darwin"):
            if not force:
                output = _subprocess.check_output(["ps", "-A"]).decode()
                processes = output.splitlines()[1:]  # EXCLUDE HEADER
                if len(processes) > 2:  # EXCLUDING THE PYTHON PROCESS AND PS
                    raise RuntimeError("Processes are still running. Use the parameter `force=True` to restart anyway.")
            if prompt:
                _subprocess.Popen(["notify-send", "System Restart", prompt])
                _time.sleep(wait)
            try:
                _subprocess.run(["sudo", "shutdown", "-r", "now"])
            except _subprocess.CalledProcessError:
                raise PermissionError("Failed to restart: insufficient privileges. Ensure sudo permissions are granted.")
            if continue_program:
                print(f"Restarting in {wait} seconds...")
                _time.sleep(wait)
        else:
            raise NotImplementedError(f"Restart not implemented for `{system}`")

    @staticmethod
    def check_libs(
        lib_names: list[str],
        install_missing: bool = False,
        confirm_install: bool = True,
    ) -> None | list[str]:
        """Checks if the given list of libraries are installed. If not:
        - If `install_missing` is `False` the missing libraries will be returned as a list.
        - If `install_missing` is `True` the missing libraries will be installed. If `confirm_install` is `True` the user will first be asked if they want to install the missing libraries.
        """
        missing = []
        for lib in lib_names:
            try:
                __import__(lib)
            except ImportError:
                missing.append(lib)
        if not missing:
            return None
        elif not install_missing:
            return missing
        if confirm_install:
            print("The following required libraries are missing:")
            for lib in missing:
                print(f"- {lib}")
            if input("Do you want to install them now (Y/n):  ").strip().lower() not in ("", "y", "yes"):
                raise ImportError("Missing required libraries.")
        try:
            _subprocess.check_call([_sys.executable, "-m", "pip", "install"] + missing)
            return None
        except _subprocess.CalledProcessError:
            return missing



from .xx_string import String
from .xx_path import Path

import os as _os


class File:

    @staticmethod
    def _make_path(
        filename: str,
        filetype: str,
        search_in: str | list[str] = None,
        prefer_base_dir: bool = True,
        correct_path: bool = False,
    ) -> str:
        """Get the path to a file in the cwd, the base-dir, or predefined directories.\n
        --------------------------------------------------------------------------------------
        If the `filename` is not found in the above directories, it will be searched<br>
        in the `search_in` directory/directories. If the file is still not found, it will<br>
        return the path to the file in the base-dir per default or to the file in the<br>
        cwd if `prefer_base_dir` is set to `False`."""
        if not filename.lower().endswith(f".{filetype.lower()}"):
            filename = f"{filename}.{filetype.lower()}"
        try:
            return Path.extend(filename, search_in, True, correct_path)
        except FileNotFoundError:
            return (
                _os.path.join(Path.get(base_dir=True), filename) if prefer_base_dir else _os.path.join(_os.getcwd(), filename)
            )

    @staticmethod
    def rename_extension(file_path: str, new_extension: str) -> str:
        directory, filename_with_ext = _os.path.split(file_path)
        filename = filename_with_ext.split(".")[0]
        camel_case_filename = String.to_camel_case(filename)
        new_filename = f"{camel_case_filename}{new_extension}"
        new_file_path = _os.path.join(directory, new_filename)
        return new_file_path

    @staticmethod
    def create(content: str = "", file: str = "new_file.txt", force: bool = False) -> str:
        """Create a file with ot without content.\n
        ----------------------------------------------------------------------------
        The function will throw a `FileExistsError` if the file already exists.<br>
        To overwrite the file, set the `force` parameter to `True`."""
        if _os.path.exists(file) and not force:
            with open(file, "r", encoding="utf-8") as existing_file:
                existing_content = existing_file.read()
                if existing_content == content:
                    raise FileExistsError("Already created this file. (nothing changed)")
            raise FileExistsError("File already exists.")
        with open(file, "w", encoding="utf-8") as f:
            f.write(content)
        full_path = _os.path.abspath(file)
        return full_path



import re as _re


class String:

    @staticmethod
    def to_type(string: str) -> any:
        """Will convert a string to a type."""
        if string.lower() in ("true", "false"):  # BOOLEAN
            return string.lower() == "true"
        elif string.lower() in ("none", "null", "undefined"):  # NONE
            return None
        elif string.startswith("[") and string.endswith("]"):  # LIST
            return [String.to_type(item.strip()) for item in string[1:-1].split(",") if item.strip()]
        elif string.startswith("(") and string.endswith(")"):  # TUPLE
            return tuple(String.to_type(item.strip()) for item in string[1:-1].split(",") if item.strip())
        elif string.startswith("{") and string.endswith("}"):  # SET
            return {String.to_type(item.strip()) for item in string[1:-1].split(",") if item.strip()}
        elif string.startswith("{") and string.endswith("}") and ":" in string:  # DICTIONARY
            return {
                String.to_type(k.strip()): String.to_type(v.strip())
                for k, v in (item.split(":") for item in string[1:-1].split(",") if item.strip())
            }
        try:  # NUMBER (INT OR FLOAT)
            if "." in string or "e" in string.lower():
                return float(string)
            else:
                return int(string)
        except ValueError:
            pass
        if string.startswith(("'", '"')) and string.endswith(("'", '"')):  # STRING (WITH OR WITHOUT QUOTES)
            return string[1:-1]
        try:  # COMPLEX
            return complex(string)
        except ValueError:
            pass
        return string  # IF NOTHING ELSE MATCHES, RETURN AS IS

    @staticmethod
    def normalize_spaces(string: str, tab_spaces: int = 4) -> str:
        """Replaces all special space characters with normal spaces.<br>
        Also replaces tab characters with `tab_spaces` spaces."""
        return (
            string.replace("\t", " " * tab_spaces)
            .replace("\u2000", " ")
            .replace("\u2001", " ")
            .replace("\u2002", " ")
            .replace("\u2003", " ")
            .replace("\u2004", " ")
            .replace("\u2005", " ")
            .replace("\u2006", " ")
            .replace("\u2007", " ")
            .replace("\u2008", " ")
            .replace("\u2009", " ")
            .replace("\u200A", " ")
        )

    @staticmethod
    def escape(string: str, str_quotes: str = '"') -> str:
        """Escapes the special characters and quotes inside a string.\n
        ----------------------------------------------------------------------------
        `str_quotes` can be either `"` or `'` and should match the quotes,<br>
        the string will be put inside of. So if your string will be `"string"`,<br>
        you should pass `"` to the parameter `str_quotes`.<br>
        That way, if the string includes the same quotes, they will be escaped."""
        string = (
            string.replace("\\", r"\\")
            .replace("\n", r"\n")
            .replace("\r", r"\r")
            .replace("\t", r"\t")
            .replace("\b", r"\b")
            .replace("\f", r"\f")
            .replace("\a", r"\a")
        )
        if str_quotes == '"':
            string = string.replace(r"\\'", "'").replace(r'"', r"\"")
        elif str_quotes == "'":
            string = string.replace(r'\\"', '"').replace(r"'", r"\'")
        return string

    @staticmethod
    def is_empty(string: str, spaces_are_empty: bool = False):
        """Returns `True` if the string is empty and `False` otherwise.<br>
        If `spaces_are_empty` is true, it will also return `True` if the string is only spaces."""
        return (string in (None, "")) or (spaces_are_empty and isinstance(string, str) and not string.strip())

    @staticmethod
    def single_char_repeats(string: str, char: str) -> int | bool:
        """If the string consists of only the same `char`, it returns the number of times it is present.<br>
        If the string doesn't consist of only the same character, it returns `False`.
        """
        if len(string) == len(char) * string.count(char):
            return string.count(char)
        else:
            return False

    @staticmethod
    def decompose(case_string: str, seps: str = "-_", lower_all: bool = True) -> list[str]:
        """Will decompose the string (*any type of casing, also mixed*) into parts."""
        return [(part.lower() if lower_all else part) for part in _re.split(rf"(?<=[a-z])(?=[A-Z])|[{seps}]", case_string)]

    @staticmethod
    def to_camel_case(string: str) -> str:
        """Will convert the string of any type of casing to camel case."""
        return "".join(part.capitalize() for part in String.decompose(string))

    @staticmethod
    def to_snake_case(string: str, sep: str = "_", screaming: bool = False) -> str:
        """Will convert the string of any type of casing to snake case."""
        return sep.join(part.upper() if screaming else part for part in String.decompose(string))

    @staticmethod
    def get_string_lines(string: str, remove_empty_lines: bool = False) -> list[str]:
        """Will split the string into lines."""
        if not remove_empty_lines:
            return string.splitlines()
        lines = string.splitlines()
        if not lines:
            return []
        non_empty_lines = [line for line in lines if line.strip()]
        if not non_empty_lines:
            return []
        return non_empty_lines

    @staticmethod
    def remove_consecutive_empty_lines(string: str, max_consecutive: int = 0) -> str:
        """Will remove consecutive empty lines from the string.\n
        ----------------------------------------------------------------------------------------------
        If `max_consecutive` is `0`, it will remove all consecutive empty lines.<br>
        If `max_consecutive` is bigger than `0`, it will only allow `max_consecutive` consecutive<br>
        empty lines and everything above it will be cut down to `max_consecutive` empty lines.
        """
        return _re.sub(r"(\n\s*){2,}", r"\1" * (max_consecutive + 1), string)

    @staticmethod
    def split_count(string: str, count: int) -> list[str]:
        """Will split the string every `count` characters."""
        return [string[i : i + count] for i in range(0, len(string), count)]

    @staticmethod
    def multi_strip(string: str, strip_chars: str = " _-") -> str:
        """Will remove all leading and trailing `strip_chars` from the string."""
        for char in string:
            if char in strip_chars:
                string = string[1:]
            else:
                break
        for char in string[::-1]:
            if char in strip_chars:
                string = string[:-1]
            else:
                break
        return string

    @staticmethod
    def multi_lstrip(string: str, strip_chars: str = " _-") -> str:
        """Will remove all leading `strip_chars` from the string."""
        for char in string:
            if char in strip_chars:
                string = string[1:]
            else:
                break
        return string

    @staticmethod
    def multi_rstrip(string: str, strip_chars: str = " _-") -> str:
        """Will remove all trailing `strip_chars` from the string."""
        for char in string[::-1]:
            if char in strip_chars:
                string = string[:-1]
            else:
                break
        return string
